@@ -1,9 +1,9 @@
 //
-// "$Id: Fl_Function_Type.cxx,v 1.15.2.16.2.15 2004/04/11 04:38:55 easysw Exp $"
+// "$Id: Fl_Function_Type.cxx 4638 2005-11-04 15:16:24Z matt $"
 //
 // C function type code for the Fast Light Tool Kit (FLTK).
 //
-// Copyright 1998-2004 by Bill Spitzak and others.
+// Copyright 1998-2005 by Bill Spitzak and others.
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Library General Public
@@ -20,10 +20,14 @@
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 // USA.
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+// Please report all bugs and problems on the following page:
+//
+//     http://www.fltk.org/str.php
 //
 
 #include <FL/Fl.H>
+#include <FL/Fl_Preferences.H>
+#include <FL/Fl_File_Chooser.H>
 #include "Fl_Type.h"
 #include <FL/fl_show_input.H>
 #include "../src/flstring.h"
@@ -177,12 +181,20 @@ void Fl_Function_Type::open() {
     if (*c && *d != '(') {
       message = "must be name(arguments), try again:"; continue;
     }
+    int mod = 0;
     c = f_return_type_input->value();
     message = c_check(c); if (message) continue;
     name(f_name_input->value());
     storestring(c, return_type);
-    public_ = f_public_button->value();
-    cdecl_ = f_c_button->value();
+    if (public_ != f_public_button->value()) {
+      mod = 1;
+      public_ = f_public_button->value();
+    }
+    if (cdecl_ != f_c_button->value()) {
+      mod = 1;
+      cdecl_ = f_c_button->value();
+    }
+    if (mod) set_modflag(1);
     break;
   }
  BREAK2:
@@ -291,7 +303,35 @@ void Fl_Function_Type::write_code1() {
 	  write_h("%s%s %s;\n", rtype, star, name());
       }
       else write_c("static ");
-      write_c("%s%s %s {\n", rtype, star, name());
+
+      // write everything but the default parameters (if any)
+      char s[1024], *sptr;
+      char *nptr;
+      int skips=0,skipc=0;
+      int nc=0,plevel=0;
+      for (sptr=s,nptr=(char*)name(); *nptr; nc++,nptr++) {
+	if (!skips && *nptr=='(') plevel++;
+	else if (!skips && *nptr==')') plevel--;
+	if ( *nptr=='"' &&  !(nc &&  *(nptr-1)=='\\') ) 
+	  skips = skips ? 0 : 1;
+	else if(!skips && *nptr=='\'' &&  !(nc &&  *(nptr-1)=='\\'))
+	  skipc = skipc ? 0 : 1;
+	if(!skips && !skipc && plevel==1 && *nptr =='=' && 
+	   !(nc && *(nptr-1)=='\'') ) // ignore '=' case 
+	  while(*++nptr  && (skips || skipc || (*nptr!=',' && *nptr!=')' || plevel!=1) )) {
+	    if ( *nptr=='"' &&  *(nptr-1)!='\\' ) 
+	      skips = skips ? 0 : 1;
+	    else if(!skips && *nptr=='\'' &&  *(nptr-1)!='\\')
+	      skipc = skipc ? 0 : 1;
+	    if (!skips && !skipc && *nptr=='(') plevel++;
+	    else if (!skips && *nptr==')') plevel--;
+	  }
+
+	if (sptr < (s + sizeof(s) - 1))	*sptr++ = *nptr;
+      }
+      *sptr = '\0';
+ 
+      write_c("%s%s %s {\n", rtype, star, s);
     }
   }
   if (havewidgets) write_c("  %s* w;\n",subclassname(child));
@@ -308,6 +348,16 @@ void Fl_Function_Type::write_code2() {
   indentation = 0;
 }
 
+int Fl_Function_Type::has_signature(const char *rtype, const char *sig) const {
+  if (!return_type) return 0;
+  if (!name()) return 0;
+  if (   strcmp(return_type, rtype)==0 
+      && fl_filename_match(name(), sig)) {
+    return 1;
+  }
+  return 0;
+}
+
 ////////////////////////////////////////////////////////////////
 
 Fl_Type *Fl_Code_Type::make() {
@@ -435,7 +485,7 @@ void Fl_CodeBlock_Type::write_code2() {
 
 int Fl_Decl_Type::is_public() const 
 {
- Fl_Type *p = Fl_Type::current;
+ Fl_Type *p = parent;
  while (p && !p->is_decl_block()) p = p->parent;
  if(p && p->is_public() && public_)
    return public_;
@@ -487,7 +537,10 @@ void Fl_Decl_Type::open() {
     message = c_check(c&&c[0]=='#' ? c+1 : c);
     if (message) continue;
     name(c);
-    public_ = decl_public_button->value();
+    if (public_!=decl_public_button->value()) {
+      set_modflag(1);
+      public_ = decl_public_button->value();
+    }
     break;
   }
  BREAK2:
@@ -519,13 +572,13 @@ void Fl_Decl_Type::write_code1() {
   while (e>c && e[-1]==';') e--;
   if (class_name(1)) {
     write_public(public_);
-    write_h("  %.*s;\n", e-c, c);
+    write_h("  %.*s;\n", (int)(e-c), c);
   } else {
     if (public_) {
-      write_h("extern %.*s;\n", e-c, c);
-      write_c("%.*s;\n", e-c, c);
+      write_h("extern %.*s;\n", (int)(e-c), c);
+      write_c("%.*s;\n", (int)(e-c), c);
     } else {
-      write_c("static %.*s;\n", e-c, c);
+      write_c("static %.*s;\n", (int)(e-c), c);
     }
   }
 }
@@ -590,7 +643,10 @@ void Fl_DeclBlock_Type::open() {
     message = c_check(c&&c[0]=='#' ? c+1 : c);
     if (message) continue;
     storestring(c,after);
-    public_ = declblock_public_button->value();
+    if (public_ != declblock_public_button->value()) {
+      set_modflag(1);
+      public_ = declblock_public_button->value();
+    }
     break;
   }
  BREAK2:
@@ -615,6 +671,241 @@ void Fl_DeclBlock_Type::write_code2() {
 
 ////////////////////////////////////////////////////////////////
 
+Fl_Type *Fl_Comment_Type::make() {
+  Fl_Type *p = Fl_Type::current;
+  while (p && !p->is_code_block()) p = p->parent;
+  Fl_Comment_Type *o = new Fl_Comment_Type();
+  o->in_c_ = 1;
+  o->in_h_ = 1;
+  o->style_ = 0;
+  o->name("my comment");
+  o->add(p);
+  o->factory = this;
+  o->title_buf[0] = 0;
+  return o;
+}
+
+void Fl_Comment_Type::write_properties() {
+  Fl_Type::write_properties();
+  if (in_c_) write_string("in_source"); else write_string("not_in_source"); 
+  if (in_h_) write_string("in_header"); else write_string("not_in_header");
+}
+
+void Fl_Comment_Type::read_property(const char *c) {
+  if (!strcmp(c,"in_source")) {
+    in_c_ = 1;
+  } else if (!strcmp(c,"not_in_source")) {
+    in_c_ = 0;
+  } else if (!strcmp(c,"in_header")) {
+    in_h_ = 1;
+  } else if (!strcmp(c,"not_in_header")) {
+    in_h_ = 0;
+  } else {
+    Fl_Type::read_property(c);
+  }
+}
+
+#include "comments.h"
+
+static void load_comments_preset(Fl_Preferences &menu) {
+  static const char * const predefined_comment[] = {
+    "GNU Public License/GPL Header",  "GNU Public License/GPL Footer",
+    "GNU Public License/LGPL Header", "GNU Public License/LGPL Footer",
+    "FLTK/Header", "FLTK/Footer" };
+  int i;
+  menu.set("n", 6);
+  Fl_Preferences db(Fl_Preferences::USER, "fltk.org", "fluid_comments");
+  for (i=0; i<6; i++) {
+    menu.set(Fl_Preferences::Name(i), predefined_comment[i]);
+    db.set(predefined_comment[i], comment_text[i]);
+  }
+}
+
+void Fl_Comment_Type::open() {
+  if (!comment_panel) make_comment_panel();
+  const char *text = name();
+  {
+    int i=0, n=0;
+    Fl_Preferences menu(Fl_Preferences::USER, "fltk.org", "fluid_comments_menu");
+    comment_predefined->clear();
+    comment_predefined->add("_Edit/Add current comment...");
+    comment_predefined->add("_Edit/Remove last selection...");
+    menu.get("n", n, -1);
+    if (n==-1) load_comments_preset(menu);
+    menu.get("n", n, 0);
+    for (i=0;i<n;i++) {
+      char *text;
+      menu.get(Fl_Preferences::Name(i), text, "");
+      comment_predefined->add(text);
+      free(text);
+    }
+  }
+  comment_input->buffer()->text( text ? text : "" );
+  comment_in_source->value(in_c_);
+  comment_in_header->value(in_h_);
+  comment_panel->show();
+  const char* message = 0;
+  char itempath[256]; itempath[0] = 0;
+  int last_selected_item = 0;
+  for (;;) { // repeat as long as there are errors
+    if (message) fl_alert(message);
+    for (;;) {
+      Fl_Widget* w = Fl::readqueue();
+      if (w == comment_panel_cancel) goto BREAK2;
+      else if (w == comment_panel_ok) break;
+      else if (w == comment_predefined) {
+        if (comment_predefined->value()==1) {
+          // add the current comment to the database
+          const char *xname = fl_input(
+            "Please enter a name to reference the current\ncomment in your database.\n\n"
+            "Use forward slashes '/' to create submenus.", 
+            "My Comment");
+          if (xname) {
+            char *name = strdup(xname);
+            for (char*s=name;*s;s++) if (*s==':') *s = ';';
+            int n;
+            Fl_Preferences db(Fl_Preferences::USER, "fltk.org", "fluid_comments");
+            db.set(name, comment_input->buffer()->text());
+            Fl_Preferences menu(Fl_Preferences::USER, "fltk.org", "fluid_comments_menu");
+            menu.get("n", n, 0);
+            menu.set(Fl_Preferences::Name(n), name);
+            menu.set("n", ++n);
+            comment_predefined->add(name);
+            free(name);
+          }
+        } else if (comment_predefined->value()==2) {
+          // remove the last selected comment from the database
+          if (itempath[0]==0 || last_selected_item==0) {
+            fl_message("Please select an entry form this menu first.");
+          } else if (fl_choice("Are you sure that you want to delete the entry\n"
+	                       "\"%s\"\nfrom the database?", "Cancel", "Delete",
+			       NULL, itempath)) {
+            Fl_Preferences db(Fl_Preferences::USER, "fltk.org", "fluid_comments");
+            db.deleteEntry(itempath);
+            comment_predefined->remove(last_selected_item);
+            Fl_Preferences menu(Fl_Preferences::USER, "fltk.org", "fluid_comments_menu");
+            int i, n;
+            for (i=4, n=0; i<comment_predefined->size(); i++) {
+              const Fl_Menu_Item *mi = comment_predefined->menu()+i;
+              if (comment_predefined->item_pathname(itempath, 255, mi)==0) {
+                if (itempath[0]=='/') memmove(itempath, itempath+1, 255);
+                if (itempath[0]) menu.set(Fl_Preferences::Name(n++), itempath);
+              }
+            }
+            menu.set("n", n);
+          }
+        } else {
+          // load the selected comment from the database
+          if (comment_predefined->item_pathname(itempath, 255)==0) {
+            if (itempath[0]=='/') memmove(itempath, itempath+1, 255);
+            Fl_Preferences db(Fl_Preferences::USER, "fltk.org", "fluid_comments");
+            char *text; 
+            db.get(itempath, text, "(no text found in data base)");
+            comment_input->buffer()->text(text);
+            free(text);
+            last_selected_item = comment_predefined->value();
+          }
+        }
+      }
+      else if (w == comment_load) {
+        // load a comment from disk
+	fl_file_chooser_ok_label("Use File");
+        const char *fname = fl_file_chooser("Pick a comment", 0L, 0L);
+	fl_file_chooser_ok_label(NULL);
+        if (fname) {
+          if (comment_input->buffer()->loadfile(fname)) {
+            fl_alert("Error loading file\n%s", fname);
+          }
+        }
+      }
+      else if (!w) Fl::wait();
+    }
+    char*c = comment_input->buffer()->text();
+    name(c);
+    free(c);
+    int mod = 0;
+    if (in_c_ != comment_in_source->value()) {
+      in_c_ = comment_in_source->value();
+      mod = 1;
+    }
+    if (in_h_ != comment_in_header->value()) {
+      in_h_ = comment_in_header->value();
+      mod = 1;
+    }
+    if (mod) set_modflag(1);
+    break;
+  }
+ BREAK2:
+  title_buf[0] = 0;
+  comment_panel->hide();
+}
+
+const char *Fl_Comment_Type::title() {
+  const char* n = name(); 
+  if (!n || !*n) return type_name();
+  if (title_buf[0]==0) {
+    const char *s = n;
+    char *d = title_buf;
+    int i = 50;
+    while (--i > 0) {
+      char n = *s++;
+      if (n==0) break;
+      if (n=='\r') { *d++ = '\\'; *d++ = 'r'; i--; }
+      else if (n=='\n') { *d++ = '\\'; *d++ = 'n'; i--; }
+      else if (n<32) { *d++ = '^'; *d++ = 'A'+n; i--; }
+      else *d++ = n;
+    }
+    if (i<=0) { *d++ = '.'; *d++ = '.'; *d++ = '.'; }
+    *d++ = 0;
+  }
+  return title_buf;
+}
+
+Fl_Comment_Type Fl_Comment_type;
+
+void Fl_Comment_Type::write_code1() {
+  const char* c = name();
+  if (!c) return;
+  if (!in_c_ && !in_h_) return;
+  // find out if there is already a valid comment:
+  const char *s = c;
+  while (isspace(*s)) s++;
+  // if this seems to be a C style comment, copy the block as is
+  // (it's up to the user to correctly close the comment)
+  if (s[0]=='/' && s[1]=='*') {
+    if (in_h_) write_h("%s\n", c);
+    if (in_c_) write_c("%s\n", c);
+    return;
+  }
+  // copy the comment line by line, add the double slash if needed
+  char *txt = strdup(c);
+  char *b = txt, *e = txt;
+  for (;;) {
+    // find the end of the line and set it to NUL
+    while (*e && *e!='\n') e++;
+    char eol = *e;
+    *e = 0;
+    // check if there is a C++ style comment at the beginning of the line
+    char *s = b;
+    while (isspace(*s)) s++;
+    if (s!=e && ( s[0]!='/' || s[1]!='/') ) {
+      // if no comment marker was found, we add one ourselves
+      if (in_h_) write_h("// ");
+      if (in_c_) write_c("// ");
+    }
+    // now copy the rest of the line
+    if (in_h_) write_h("%s\n", b);
+    if (in_c_) write_c("%s\n", b);
+    if (eol==0) break;
+    *e++ = eol;
+    b = e;
+  }
+}
+
+void Fl_Comment_Type::write_code2() {}
+
+////////////////////////////////////////////////////////////////
+
 const char* Fl_Type::class_name(const int need_nest) const {
   Fl_Type* p = parent;
   while (p) {
@@ -637,6 +928,20 @@ const char* Fl_Type::class_name(const int need_nest) const {
   return 0;
 }
 
+/**
+ * If this Type resides inside a class, this function returns the class type, or null.
+ */
+const Fl_Class_Type *Fl_Type::is_in_class() const {
+  Fl_Type* p = parent;
+  while (p) {
+    if (p->is_class()) {
+      return (Fl_Class_Type*)p;
+    }
+    p = p->parent;
+  }
+  return 0;
+}
+
 int Fl_Class_Type::is_public() const {return public_;}
 
 void Fl_Class_Type::prefix(const char*p) {
@@ -725,7 +1030,10 @@ void Fl_Class_Type::open() {
     prefix(pr);
     free((void*)s);
     storestring(c, subclass_of);
-    public_ = c_public_button->value();
+    if (public_ != c_public_button->value()) {
+      public_ = c_public_button->value();
+      set_modflag(1);
+    }
     break;
   }
  BREAK2:
@@ -735,11 +1043,14 @@ void Fl_Class_Type::open() {
 Fl_Class_Type Fl_Class_type;
 
 static Fl_Class_Type *current_class;
+extern Fl_Widget_Class_Type *current_widget_class;
 extern int varused_test;
 void write_public(int state) {
-  if (!current_class || varused_test) return;
-  if (current_class->write_public_state == state) return;
-  current_class->write_public_state = state;
+  if ((!current_class && !current_widget_class) || varused_test) return;
+  if (current_class && current_class->write_public_state == state) return;
+  if (current_widget_class && current_widget_class->write_public_state == state) return;
+  if (current_class) current_class->write_public_state = state;
+  if (current_widget_class) current_widget_class->write_public_state = state;
   write_h(state ? "public:\n" : "private:\n");
 }
 
@@ -760,6 +1071,21 @@ void Fl_Class_Type::write_code2() {
   current_class = parent_class;
 }
 
+/**
+ * Return 1 if this class contains a function with the given signature.
+ */
+int Fl_Class_Type::has_function(const char *rtype, const char *sig) const {
+  Fl_Type *child;
+  for (child = next; child && child->level > level; child = child->next) {
+    if (child->level == level+1 && strcmp(child->type_name(), "Function")==0) {
+      const Fl_Function_Type *fn = (const Fl_Function_Type*)child;
+      if (fn->has_signature(rtype, sig))
+        return 1;
+    }
+  }
+  return 0;
+}
+
 //
-// End of "$Id: Fl_Function_Type.cxx,v 1.15.2.16.2.15 2004/04/11 04:38:55 easysw Exp $".
+// End of "$Id: Fl_Function_Type.cxx 4638 2005-11-04 15:16:24Z matt $".
 //
