@@ -1,9 +1,9 @@
 //
-// "$Id: Fl_Menu.cxx,v 1.18.2.12.2.34 2004/10/18 20:22:23 easysw Exp $"
+// "$Id: Fl_Menu.cxx 4766 2006-01-17 21:04:55Z mike $"
 //
 // Menu code for the Fast Light Tool Kit (FLTK).
 //
-// Copyright 1998-2004 by Bill Spitzak and others.
+// Copyright 1998-2005 by Bill Spitzak and others.
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Library General Public
@@ -20,7 +20,9 @@
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 // USA.
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+// Please report all bugs and problems on the following page:
+//
+//     http://www.fltk.org/str.php
 //
 
 // Warning: this menu code is quite a mess!
@@ -35,6 +37,10 @@
 #include <FL/fl_draw.H>
 #include <stdio.h>
 
+#ifdef __APPLE__
+#  include <Carbon/Carbon.h>
+#endif
+
 int Fl_Menu_Item::size() const {
   const Fl_Menu_Item* m = this;
   int nest = 0;
@@ -116,9 +122,10 @@ int Fl_Menu_Item::measure(int* hp, const Fl_Menu_* m) const {
   l.type    = labeltype_;
   l.font    = labelsize_ || labelfont_ ? labelfont_ : uchar(m ? m->textfont() : FL_HELVETICA);
   l.size    = labelsize_ ? labelsize_ : m ? m->textsize() : (uchar)FL_NORMAL_SIZE;
-  l.color   = FL_BLACK; // this makes no difference?
+  l.color   = FL_FOREGROUND_COLOR; // this makes no difference?
   fl_draw_shortcut = 1;
-  int w = 0; int h = 0; l.measure(w, hp ? *hp : h);
+  int w = 0; int h = 0;
+  l.measure(w, hp ? *hp : h);
   fl_draw_shortcut = 0;
   if (flags & (FL_MENU_TOGGLE|FL_MENU_RADIO)) w += 14;
   return w;
@@ -133,7 +140,7 @@ void Fl_Menu_Item::draw(int x, int y, int w, int h, const Fl_Menu_* m,
   l.type    = labeltype_;
   l.font    = labelsize_ || labelfont_ ? labelfont_ : uchar(m ? m->textfont() : FL_HELVETICA);
   l.size    = labelsize_ ? labelsize_ : m ? m->textsize() : (uchar)FL_NORMAL_SIZE;
-  l.color   = labelcolor_ ? labelcolor_ : m ? m->textcolor() : int(FL_BLACK);
+  l.color   = labelcolor_ ? labelcolor_ : m ? m->textcolor() : int(FL_FOREGROUND_COLOR);
   if (!active()) l.color = fl_inactive((Fl_Color)l.color);
   Fl_Color color = m ? m->color() : FL_GRAY;
   if (selected) {
@@ -234,7 +241,11 @@ menuwindow::menuwindow(const Fl_Menu_Item* m, int X, int Y, int Wp, int Hp,
 		       int menubar, int menubar_title, int right_edge)
   : Fl_Menu_Window(X, Y, Wp, Hp, 0)
 {
-  if (!right_edge) right_edge = Fl::w();
+  int scr_x, scr_y, scr_w, scr_h;
+  int tx = X, ty = Y;
+
+  Fl::screen_xywh(scr_x, scr_y, scr_w, scr_h);
+  if (!right_edge || right_edge > scr_x+scr_w) right_edge = scr_x+scr_w;
 
   end();
   set_modal();
@@ -290,24 +301,41 @@ menuwindow::menuwindow(const Fl_Menu_Item* m, int X, int Y, int Wp, int Hp,
   if (Wp > W) W = Wp;
   if (Wtitle > W) W = Wtitle;
 
-  if (!Wp) {if (X < 0) X = 0; if (X > Fl::w()-W) X= right_edge-W;}
+  if (X < scr_x) X = scr_x; if (X > scr_x+scr_w-W) X= scr_x+scr_w-W;
   x(X); w(W);
   h((numitems ? itemheight*numitems-LEADING : 0)+2*BW+3);
   if (selected >= 0)
     Y = Y+(Hp-itemheight)/2-selected*itemheight-BW;
-  else
+  else {
     Y = Y+Hp;
+    // if the menu hits the bottom of the screen, we try to draw
+    // it above the menubar instead. We will not adjust any menu
+    // that has a selected item.
+    if (Y+h()>scr_y+scr_h && Y-h()>=scr_y) {
+      if (Hp>1) 
+        // if we know the height of the Fl_Menu_, use it
+        Y = Y-Hp-h();
+      else if (t)
+        // assume that the menubar item height relates to the first
+        // menuitem as well
+        Y = Y-itemheight-h()-Fl::box_dh(box());
+      else
+        // draw the menu to the right
+        Y = Y-h()+itemheight+Fl::box_dy(box());
+    }
+  }
   if (m) y(Y); else {y(Y-2); w(1); h(1);}
 
   if (t) {
     int dy = menubar_title ? Fl::box_dy(button->box())+1 : 2;
     int ht = menubar_title ? button->h()-dy*2 : Htitle+2*BW+3;
-    title = new menutitle(X, Y-ht-dy, Wtitle, ht, t);
+    title = new menutitle(tx, ty-ht-dy, Wtitle, ht, t);
   } else
     title = 0;
 }
 
 menuwindow::~menuwindow() {
+  hide();
   delete title;
 }
 
@@ -319,10 +347,13 @@ void menuwindow::position(int X, int Y) {
 
 // scroll so item i is visible on screen
 void menuwindow::autoscroll(int n) {
+  int scr_x, scr_y, scr_w, scr_h;
   int Y = y()+Fl::box_dx(box())+2+n*itemheight;
-  if (Y <= Fl::y()) Y = Fl::y()-Y+10;
+
+  Fl::screen_xywh(scr_x, scr_y, scr_w, scr_h);
+  if (Y <= scr_y) Y = scr_y-Y+10;
   else {
-    Y = Y+itemheight-Fl::h()-Fl::y();
+    Y = Y+itemheight-scr_h-scr_y;
     if (Y < 0) return;
     Y = -Y-10;
   }
@@ -352,10 +383,10 @@ void menuwindow::drawentry(const Fl_Menu_Item* m, int n, int eraseit) {
 
   // the shortcuts and arrows assumme fl_color() was left set by draw():
   if (m->submenu()) {
-    int sz = (hh-5)&-2;
+    int sz = (hh-7)&-2;
     int y1 = yy+(hh-sz)/2;
     int x1 = xx+ww-sz-3;
-    fl_polygon(x1, y1, x1, y1+sz, x1+sz, y1+sz/2);
+    fl_polygon(x1+2, y1, x1+2, y1+sz, x1+sz/2+2, y1+sz/2);
   } else if (m->shortcut_) {
     Fl_Font f = m->labelsize_ || m->labelfont_ ? (Fl_Font)m->labelfont_ :
                     button ? button->textfont() : FL_HELVETICA;
@@ -556,13 +587,28 @@ int menuwindow::handle(int e) {
   case FL_MOVE:
   case FL_PUSH:
   case FL_DRAG: {
+#ifdef __QNX__
+    // STR 704: workaround QNX X11 bug - in QNX a FL_MOVE event is sent
+    // right after FL_RELEASE...
+    if (pp.state == DONE_STATE) return 1;
+#endif // __QNX__
     int mx = Fl::event_x_root();
     int my = Fl::event_y_root();
-    int item=0; int mymenu;
+    int item=0; int mymenu = pp.nummenus-1;
+    if (e == FL_PUSH && (!pp.menubar || mymenu) &&
+        (mx < pp.p[mymenu]->x_root() ||
+	 mx >= (pp.p[mymenu]->x_root() + pp.p[mymenu]->w()) ||
+         my < pp.p[mymenu]->y_root() ||
+	 my >= (pp.p[mymenu]->y_root() + pp.p[mymenu]->h()))) {
+      // Clicking outside menu cancels it...
+      setitem(0, -1, 0);
+      pp.state = DONE_STATE;
+      return 1;
+    }
     for (mymenu = pp.nummenus-1; ; mymenu--) {
       item = pp.p[mymenu]->find_selected(mx, my);
       if (item >= 0) break;
-      if (mymenu <= 0) break;
+      if (mymenu <= 0) return 0;
     }
     if (my == 0 && item > 0) setitem(mymenu, item - 1);
     else setitem(mymenu, item);
@@ -576,7 +622,13 @@ int menuwindow::handle(int e) {
     }} return 1;
   case FL_RELEASE:
     // do nothing if they try to pick inactive items
-    if (pp.current_item && !pp.current_item->activevisible()) return 1;
+    if (pp.current_item && !pp.current_item->activevisible()) {
+      if (pp.state==INITIAL_STATE) {
+        setitem(0, -1, 0);
+        pp.state = DONE_STATE;
+      }
+      return 1;
+    }
     // Mouse must either be held down/dragged some, or this must be
     // the second click (not the one that popped up the menu):
     if (!Fl::event_is_click() || pp.state == PUSH_STATE ||
@@ -632,7 +684,13 @@ const Fl_Menu_Item* Fl_Menu_Item::pulldown(
   }
 
   pp.current_item = 0; pp.menu_number = 0; pp.item_number = -1;
-  if (menubar) mw.handle(FL_DRAG); // find the initial menu
+  if (menubar) {
+    // find the initial menu
+    if (!mw.handle(FL_DRAG)) {
+      Fl::release();
+      return 0;
+    }
+  }
   initial_item = pp.current_item;
   if (initial_item) goto STARTUP;
 
@@ -733,10 +791,10 @@ const Fl_Menu_Item* Fl_Menu_Item::pulldown(
     }
   }
   const Fl_Menu_Item* m = pp.current_item;
+  Fl::release();
   delete fakemenu;
   while (pp.nummenus>1) delete pp.p[--pp.nummenus];
   mw.hide();
-  Fl::release();
   return m;
 }
 
@@ -791,5 +849,5 @@ const Fl_Menu_Item* Fl_Menu_Item::test_shortcut() const {
 }
 
 //
-// End of "$Id: Fl_Menu.cxx,v 1.18.2.12.2.34 2004/10/18 20:22:23 easysw Exp $".
+// End of "$Id: Fl_Menu.cxx 4766 2006-01-17 21:04:55Z mike $".
 //
