@@ -1,9 +1,9 @@
 //
-// "$Id: Fl_Preferences.cxx,v 1.1.2.27 2004/04/11 04:38:58 easysw Exp $"
+// "$Id: Fl_Preferences.cxx 4660 2005-11-27 14:45:48Z mike $"
 //
 // Preferences methods for the Fast Light Tool Kit (FLTK).
 //
-// Copyright 2002-2004 by Matthias Melcher.
+// Copyright 2002-2005 by Matthias Melcher.
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Library General Public
@@ -20,7 +20,9 @@
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 // USA.
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+// Please report all bugs and problems on the following page:
+//
+//     http://www.fltk.org/str.php
 //
 
 
@@ -37,8 +39,12 @@
 #if defined(WIN32) && !defined(__CYGWIN__)
 #  include <direct.h>
 #  include <io.h>
+// Visual C++ 2005 incorrectly displays a warning about the use of POSIX APIs
+// on Windows, which is supposed to be POSIX compliant...
+#  define access _access
+#  define mkdir _mkdir
 #elif defined (__APPLE__)
-#  include <Carbon/Carbon.H>
+#  include <Carbon/Carbon.h>
 #  include <unistd.h>
 #else
 #  include <unistd.h>
@@ -85,7 +91,7 @@ Fl_Preferences::Fl_Preferences( const char *path, const char *vendor, const char
  */
 Fl_Preferences::Fl_Preferences( Fl_Preferences &parent, const char *key )
 {
-  rootNode = 0;
+  rootNode = parent.rootNode;
   node = parent.node->addChild( key );
 }
 
@@ -98,7 +104,7 @@ Fl_Preferences::Fl_Preferences( Fl_Preferences &parent, const char *key )
  */
 Fl_Preferences::Fl_Preferences( Fl_Preferences *parent, const char *key )
 {
-  rootNode = 0;
+  rootNode = parent->rootNode;
   node = parent->node->addChild( key );
 }
 
@@ -110,7 +116,7 @@ Fl_Preferences::Fl_Preferences( Fl_Preferences *parent, const char *key )
  */
 Fl_Preferences::~Fl_Preferences()
 {
-  delete rootNode;
+  if (!node->parent()) delete rootNode;
   // DO NOT delete nodes! The root node will do that after writing the preferences
 }
 
@@ -616,7 +622,11 @@ Fl_Preferences::RootNode::RootNode( Fl_Preferences *prefs, Root root, const char
 	nn = FL_PATH_MAX - appDataLen;
 	err = RegQueryValueEx( key, "AppData", 0L, &type, (BYTE*)filename, &nn );
 	if ( ( err != ERROR_SUCCESS ) && ( type == REG_SZ ) )
-	  filename[0] = 0;
+	{
+	  err = RegQueryValueEx( key, "Personal", 0L, &type, (BYTE*)filename, &nn );
+	  if ( ( err != ERROR_SUCCESS ) && ( type == REG_SZ ) )
+	    filename[0] = 0;
+	}
         RegCloseKey(key);
       }
       break;
@@ -877,7 +887,7 @@ Fl_Preferences::Node *Fl_Preferences::Node::addChild( const char *path )
 {
   sprintf( nameBuffer, "%s/%s", path_, path );
   char *name = strdup( nameBuffer );
-  Node *nd = find( nameBuffer );
+  Node *nd = find( name );
   free( name );
   dirty_ = 1;
   return nd;
@@ -1105,13 +1115,13 @@ char Fl_Preferences::Node::remove()
 	break;
       }
     }
+    parent_->dirty_ = 1;
   }
   delete this;
-  dirty_ = 1;
   return ( nd != 0 );
 }
 
 
 //
-// End of "$Id: Fl_Preferences.cxx,v 1.1.2.27 2004/04/11 04:38:58 easysw Exp $".
+// End of "$Id: Fl_Preferences.cxx 4660 2005-11-27 14:45:48Z mike $".
 //
