@@ -1,9 +1,9 @@
 //
-// "$Id: Fl_mac.cxx,v 1.1.2.64 2004/09/09 21:34:46 matthiaswm Exp $"
+// "$Id: Fl_mac.cxx 4748 2006-01-15 02:26:54Z mike $"
 //
 // MacOS specific code for the Fast Light Tool Kit (FLTK).
 //
-// Copyright 1998-2004 by Bill Spitzak and others.
+// Copyright 1998-2006 by Bill Spitzak and others.
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Library General Public
@@ -20,7 +20,9 @@
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 // USA.
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+// Please report all bugs and problems on the following page:
+//
+//     http://www.fltk.org/str.php
 //
 
 //// From the inner edge of a MetroWerks CodeWarrior CD:
@@ -143,7 +145,40 @@ static unsigned short macKeyLookUp[128] =
     FL_F+2, FL_Page_Down, FL_F+1, FL_Left, FL_Right, FL_Down, FL_Up, 0,
 };
 
+/**
+ * convert the current mouse chord into the FLTK modifier state
+ */
+static void mods_to_e_state( UInt32 mods )
+{
+  long state = 0;
+  if ( mods & kEventKeyModifierNumLockMask ) state |= FL_NUM_LOCK;
+  if ( mods & cmdKey ) state |= FL_META;
+  if ( mods & (optionKey|rightOptionKey) ) state |= FL_ALT;
+  if ( mods & (controlKey|rightControlKey) ) state |= FL_CTRL;
+  if ( mods & (shiftKey|rightShiftKey) ) state |= FL_SHIFT;
+  if ( mods & alphaLock ) state |= FL_CAPS_LOCK;
+  Fl::e_state = ( Fl::e_state & 0xff000000 ) | state;
+  //printf( "State 0x%08x (%04x)\n", Fl::e_state, mods );
+}
 
+
+/**
+ * convert the current mouse chord into the FLTK keysym
+ */
+static void mods_to_e_keysym( UInt32 mods )
+{
+  if ( mods & cmdKey ) Fl::e_keysym = FL_Meta_L;
+  else if ( mods & kEventKeyModifierNumLockMask ) Fl::e_keysym = FL_Num_Lock;
+  else if ( mods & optionKey ) Fl::e_keysym = FL_Alt_L;
+  else if ( mods & rightOptionKey ) Fl::e_keysym = FL_Alt_R;
+  else if ( mods & controlKey ) Fl::e_keysym = FL_Control_L;
+  else if ( mods & rightControlKey ) Fl::e_keysym = FL_Control_R;
+  else if ( mods & shiftKey ) Fl::e_keysym = FL_Shift_L;
+  else if ( mods & rightShiftKey ) Fl::e_keysym = FL_Shift_R;
+  else if ( mods & alphaLock ) Fl::e_keysym = FL_Caps_Lock;
+  else Fl::e_keysym = 0;
+  //printf( "to sym 0x%08x (%04x)\n", Fl::e_keysym, mods );
+}
 // these pointers are set by the Fl::lock() function:
 static void nothing() {}
 void (*fl_lock_function)() = nothing;
@@ -418,8 +453,8 @@ void Fl::remove_fd(int n)
  */
 int fl_ready()
 {
-  if (GetNumEventsInQueue(GetCurrentEventQueue()) > 0) return 1;
-  return 0;
+  EventRef event;
+  return !ReceiveNextEvent(0, NULL, 0.0, false, &event);
 }
 
 /**
@@ -541,7 +576,6 @@ static pascal OSStatus carbonDispatchHandler( EventHandlerCallRef nextHandler, E
   }
   if ( ret == eventNotHandledErr )
     ret = CallNextEventHandler( nextHandler, event ); // let the OS handle the activation, but continue to get a click-through effect
-  QuitApplicationEventLoop();
 
   fl_unlock_function();
 
@@ -549,18 +583,6 @@ static pascal OSStatus carbonDispatchHandler( EventHandlerCallRef nextHandler, E
 }
 
 
-/**
- * this callback simply quits the main event loop handler, so FLTK can do its magic
- */
-static pascal void timerProcCB( EventLoopTimerRef, void* )
-{
-  fl_lock_function();
-
-  QuitApplicationEventLoop();
-
-  fl_unlock_function();
-}
-
 /**
  * break the current event loop
  */
@@ -577,6 +599,51 @@ static void breakMacEventLoop()
   fl_unlock_function();
 }
 
+//
+// MacOS X timers
+//
+
+struct MacTimeout {
+    Fl_Timeout_Handler callback;
+    void* data;
+    EventLoopTimerRef timer;
+};
+static MacTimeout* mac_timers;
+static int mac_timer_alloc;
+static int mac_timer_used;
+
+
+static void realloc_timers()
+{
+    if (mac_timer_alloc == 0) {
+        mac_timer_alloc = 8;
+    }
+    MacTimeout* new_timers = new MacTimeout[mac_timer_alloc * 2];
+    memmove(new_timers, mac_timers, sizeof(MacTimeout) * mac_timer_used);
+    MacTimeout* delete_me = mac_timers;
+    mac_timers = new_timers;
+    delete [] delete_me;
+    mac_timer_alloc *= 2;
+}
+
+static void delete_timer(MacTimeout& t)
+{
+    RemoveEventLoopTimer(t.timer);
+    memset(&t, 0, sizeof(MacTimeout));
+}
+
+
+static pascal void do_timer(EventLoopTimerRef timer, void* data)
+{
+   for (int i = 0;  i < mac_timer_used;  ++i) {
+        MacTimeout& t = mac_timers[i];
+        if (t.timer == timer  &&  t.data == data) {
+            (*t.callback)(data);
+            break;
+        }
+    }
+    breakMacEventLoop();
+}
 
 /**
  * This function is the central event handler.
@@ -599,7 +666,6 @@ static double do_queued_events( double time = 0.0 )
   }
   OSStatus ret;
   static EventTargetRef target = 0;
-  static EventLoopTimerRef timer = 0;
   if ( !target ) 
   {
     target = GetEventDispatcherTarget();
@@ -618,6 +684,7 @@ static double do_queued_events( double time = 0.0 )
         { kEventClassMouse, kEventMouseDown },
         { kEventClassMouse, kEventMouseUp },
         { kEventClassMouse, kEventMouseMoved },
+        { kEventClassMouse, 11 }, // MightyMouse wheels
         { kEventClassMouse, kEventMouseWheelMoved },
         { kEventClassMouse, kEventMouseDragged },
         { kEventClassFLTK, kEventFLTKBreakLoop },
@@ -626,7 +693,6 @@ static double do_queued_events( double time = 0.0 )
     static EventTypeSpec appEvents[] = {
         { kEventClassCommand, kEventCommandProcess } };
     ret = InstallApplicationEventHandler( dispatchHandler, GetEventTypeCount(appEvents), appEvents, 0, 0L );
-    ret = InstallEventLoopTimer( GetMainEventLoop(), 0, 0, NewEventLoopTimerUPP( timerProcCB ), 0, &timer );
   }
 
   got_events = 0;
@@ -641,10 +707,30 @@ static double do_queued_events( double time = 0.0 )
 
   fl_unlock_function();
 
-  SetEventLoopTimerNextFireTime( timer, time );
-  RunApplicationEventLoop(); // will return after the previously set time
-  if ( dataready.IsThreadRunning() )
-    { dataready.CancelThread(DEBUGTEXT("APPEVENTLOOP DONE\n")); }
+  EventRef event;
+  EventTimeout timeout = time;
+  if (!ReceiveNextEvent(0, NULL, timeout, true, &event)) {
+    got_events = 1;
+    OSErr ret = SendEventToEventTarget( event, target );
+    if (ret!=noErr) {
+      EventRecord clevent;
+      ConvertEventRefToEventRecord(event, &clevent);
+      if (clevent.what==kHighLevelEvent) {
+        ret = AEProcessAppleEvent(&clevent);
+      }
+    }
+    if (   ret==eventNotHandledErr
+        && GetEventClass(event)==kEventClassMouse
+        && GetEventKind(event)==kEventMouseDown ) {
+      WindowRef win; Point pos;
+      GetEventParameter(event, kEventParamMouseLocation, typeQDPoint,
+        NULL, sizeof(pos), NULL, &pos);
+      if (MacFindWindow(pos, &win)==inMenuBar) {
+        MenuSelect(pos);
+      }
+    }
+    ReleaseEvent( event );
+  }
 
   fl_lock_function();
 
@@ -689,7 +775,6 @@ static OSErr QuitAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* repl
       return noErr; // FLTK has not close all windows, so we return to the main program now
     }
   }
-  QuitApplicationEventLoop();
 
   fl_unlock_function();
 
@@ -746,6 +831,7 @@ static pascal OSStatus carbonWindowHandler( EventHandlerCallRef nextHandler, Eve
         activeWindow = window;
       }
       Fl::handle( FL_SHOW, window);
+      mods_to_e_state(GetCurrentKeyModifiers());
     }
     break;
   case kEventWindowHidden:
@@ -788,25 +874,28 @@ static pascal OSStatus carbonWindowHandler( EventHandlerCallRef nextHandler, Eve
  */
 static pascal OSStatus carbonMousewheelHandler( EventHandlerCallRef nextHandler, EventRef event, void *userData )
 {
-  Fl_Window *window = (Fl_Window*)userData;
-  EventMouseWheelAxis axis;
-
+  // Handle the new "MightyMouse" mouse wheel events. Please, someone explain
+  // to me why Apple changed the API on this even though the current API
+  // supports two wheels just fine. Matthias,
   fl_lock_function();
-  
+
+  fl_os_event = event;
+  Fl_Window *window = (Fl_Window*)userData;
+
+  EventMouseWheelAxis axis;
   GetEventParameter( event, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(EventMouseWheelAxis), NULL, &axis );
   long delta;
   GetEventParameter( event, kEventParamMouseWheelDelta, typeLongInteger, NULL, sizeof(long), NULL, &delta );
-  if ( axis == kEventMouseWheelAxisX )
-  {
-    Fl::e_dx = delta;
+//  fprintf(stderr, "axis=%d, delta=%d\n", axis, delta);
+  if ( axis == kEventMouseWheelAxisX ) {
+    Fl::e_dx = -delta;
+    Fl::e_dy = 0;
     if ( Fl::e_dx) Fl::handle( FL_MOUSEWHEEL, window );
-  }
-  else if ( axis == kEventMouseWheelAxisY )
-  {
+  } else if ( axis == kEventMouseWheelAxisY ) {
+    Fl::e_dx = 0;
     Fl::e_dy = -delta;
     if ( Fl::e_dy) Fl::handle( FL_MOUSEWHEEL, window );
-  }
-  else {
+  } else {
     fl_unlock_function();
 
     return eventNotHandledErr;
@@ -869,7 +958,10 @@ static pascal OSStatus carbonMouseHandler( EventHandlerCallRef nextHandler, Even
     fl_os_capture = xid;
     sendEvent = FL_PUSH;
     Fl::e_is_click = 1; px = pos.h; py = pos.v;
-    Fl::e_clicks = clickCount-1;
+    if (clickCount>1) 
+      Fl::e_clicks++;
+    else
+      Fl::e_clicks = 0;
     // fall through
   case kEventMouseUp:
     if ( !window ) break;
@@ -910,41 +1002,6 @@ static pascal OSStatus carbonMouseHandler( EventHandlerCallRef nextHandler, Even
 }
 
 
-/**
- * convert the current mouse chord into the FLTK modifier state
- */
-static void mods_to_e_state( UInt32 mods )
-{
-  long state = 0;
-  if ( mods & kEventKeyModifierNumLockMask ) state |= FL_NUM_LOCK;
-  if ( mods & cmdKey ) state |= FL_META;
-  if ( mods & (optionKey|rightOptionKey) ) state |= FL_ALT;
-  if ( mods & (controlKey|rightControlKey) ) state |= FL_CTRL;
-  if ( mods & (shiftKey|rightShiftKey) ) state |= FL_SHIFT;
-  if ( mods & alphaLock ) state |= FL_CAPS_LOCK;
-  Fl::e_state = ( Fl::e_state & 0xff000000 ) | state;
-  //printf( "State 0x%08x (%04x)\n", Fl::e_state, mods );
-}
-
-
-/**
- * convert the current mouse chord into the FLTK keysym
- */
-static void mods_to_e_keysym( UInt32 mods )
-{
-  if ( mods & cmdKey ) Fl::e_keysym = FL_Meta_L;
-  else if ( mods & kEventKeyModifierNumLockMask ) Fl::e_keysym = FL_Num_Lock;
-  else if ( mods & optionKey ) Fl::e_keysym = FL_Alt_L;
-  else if ( mods & rightOptionKey ) Fl::e_keysym = FL_Alt_R;
-  else if ( mods & controlKey ) Fl::e_keysym = FL_Control_L;
-  else if ( mods & rightControlKey ) Fl::e_keysym = FL_Control_R;
-  else if ( mods & shiftKey ) Fl::e_keysym = FL_Shift_L;
-  else if ( mods & rightShiftKey ) Fl::e_keysym = FL_Shift_R;
-  else if ( mods & alphaLock ) Fl::e_keysym = FL_Caps_Lock;
-  else Fl::e_keysym = 0;
-  //printf( "to sym 0x%08x (%04x)\n", Fl::e_keysym, mods );
-}
-
 /**
  * convert the keyboard return code into the symbol on the keycaps
  */
@@ -1335,18 +1392,70 @@ int Fl_X::fake_X_wm(const Fl_Window* w,int &X,int &Y, int &bt,int &bx, int &by)
   H = w->h()+dy;
 
   //Proceed to positioning the window fully inside the screen, if possible
-  //Make border's lower right corner visible
-  if (Fl::w() < X+W) X = Fl::w() - W;
-  if (Fl::h() < Y+H) Y = Fl::h() - H;
-  //Make border's upper left corner visible
-  if (X<0) X = 0;
-  if (Y<0) Y = 0;
-  //Make client area's lower right corner visible
-  if (Fl::w() < X+dx+ w->w()) X = Fl::w() - w->w() - dx;
-  if (Fl::h() < Y+dy+ w->h()) Y = Fl::h() - w->h() - dy;
-  //Make client area's upper left corner visible
-  if (X+xoff < 0) X = -xoff;
-  if (Y+yoff < 0) Y = -yoff;
+
+  // let's get a little elaborate here. Mac OS X puts a lot of stuff on the desk
+  // that we want to avoid when positioning our window, namely the Dock and the
+  // top menu bar (and even more stuff in 10.4 Tiger). So we will go through the
+  // list of all available screens and find the one that this window is most
+  // likely to go to, and then reposition it to fit withing the 'good' area.
+  Rect r;
+  // find the screen, that the center of this window will fall into
+  int R = X+W, B = Y+H; // right and bottom
+  int cx = (X+R)/2, cy = (Y+B)/2; // center of window;
+  GDHandle gd = 0L;
+  for (gd = GetDeviceList(); gd; gd = GetNextDevice(gd)) {
+  GDPtr gp = *gd;
+  if (    cx >= gp->gdRect.left && cx <= gp->gdRect.right
+       && cy >= gp->gdRect.top  && cy <= gp->gdRect.bottom)
+    break;
+  }
+  // if the center doesn't fall on a screen, try the top left
+  if (!gd) {
+    for (gd = GetDeviceList(); gd; gd = GetNextDevice(gd)) {
+      GDPtr gp = *gd;
+      if (    X >= gp->gdRect.left && X <= gp->gdRect.right
+           && Y >= gp->gdRect.top  && Y <= gp->gdRect.bottom)
+        break;
+    }
+  }
+  // if that doesn't fall on a screen, try the top right
+  if (!gd) {
+    for (gd = GetDeviceList(); gd; gd = GetNextDevice(gd)) {
+      GDPtr gp = *gd;
+      if (    R >= gp->gdRect.left && R <= gp->gdRect.right
+           && Y >= gp->gdRect.top  && Y <= gp->gdRect.bottom)
+        break;
+    }
+  }
+  // if that doesn't fall on a screen, try the bottom left
+  if (!gd) {
+    for (gd = GetDeviceList(); gd; gd = GetNextDevice(gd)) {
+      GDPtr gp = *gd;
+      if (    X >= gp->gdRect.left && X <= gp->gdRect.right
+           && B >= gp->gdRect.top  && B <= gp->gdRect.bottom)
+        break;
+    }
+  }
+  // last resort, try the bottom right
+  if (!gd) {
+    for (gd = GetDeviceList(); gd; gd = GetNextDevice(gd)) {
+      GDPtr gp = *gd;
+      if (    R >= gp->gdRect.left && R <= gp->gdRect.right
+           && B >= gp->gdRect.top  && B <= gp->gdRect.bottom)
+        break;
+    }
+  }
+  // if we still have not found a screen, we will use the main
+  // screen, the one that has the application menu bar.
+  if (!gd) gd = GetMainDevice();
+  if (gd) {
+    GetAvailableWindowPositioningBounds(gd, &r);
+    if ( R > r.right )  X -= R - r.right;
+    if ( B > r.bottom ) Y -= B - r.bottom;
+    if ( X < r.left )   X = r.left;
+    if ( Y < r.top )    Y = r.top;
+  }
+
   //Return the client area's top left corner in (X,Y)
   X+=xoff;
   Y+=yoff;
@@ -1364,9 +1473,93 @@ static int FSSpec2UnixPath( FSSpec *fs, char *dst )
   FSRefMakePath( &fsRef, (UInt8*)dst, 1024 );
   return strlen(dst);
 }
- 
+
+static DragReference currDragRef = 0;
+static char *currDragData = 0L;
+static int currDragSize = 0; 
+static OSErr currDragErr = noErr;
 Fl_Window *fl_dnd_target_window = 0;
 #include <FL/fl_draw.H>
+
+/**
+ * Fill the currDrag* variables with the current DnD ASCII text.
+ */
+static OSErr fillCurrentDragData(DragReference dragRef)
+{
+  OSErr ret = noErr;
+  char *dst = 0L;
+  
+  // shortcut through this whole procedure if this is still the same drag event
+  if (dragRef==currDragRef)
+    return currDragErr;
+  
+  // clear currDrag* for a new drag event
+  currDragRef = dragRef;
+  if (currDragData) free(currDragData);
+  currDragData = 0;
+  currDragSize = 0;
+  
+  // fill currDRag* with ASCII data, if available
+  UInt16 i, nItem;
+  ItemReference itemRef;
+  FlavorFlags flags;
+  Size itemSize, size = 0;
+  CountDragItems( dragRef, &nItem );
+  for ( i = 1; i <= nItem; i++ )
+  {
+    GetDragItemReferenceNumber( dragRef, i, &itemRef );
+    ret = GetFlavorFlags( dragRef, itemRef, 'TEXT', &flags );
+    if ( ret == noErr )
+    {
+      GetFlavorDataSize( dragRef, itemRef, 'TEXT', &itemSize );
+      size += itemSize;
+    }
+    ret = GetFlavorFlags( dragRef, itemRef, 'hfs ', &flags );
+    if ( ret == noErr )
+    {
+      size += 1024; //++ ouch! We should create the full pathname and figure out its length
+    }
+  }
+
+  if ( !size )
+  {
+    currDragErr = userCanceledErr;
+    return currDragErr;
+  }
+
+  currDragSize = size + nItem - 1;
+  currDragData = dst = (char*)malloc( size+nItem );;
+
+  for ( i = 1; i <= nItem; i++ )
+  {
+    GetDragItemReferenceNumber( dragRef, i, &itemRef );
+    ret = GetFlavorFlags( dragRef, itemRef, 'TEXT', &flags );
+    if ( ret == noErr )
+    {
+      GetFlavorDataSize( dragRef, itemRef, 'TEXT', &itemSize );
+      GetFlavorData( dragRef, itemRef, 'TEXT', dst, &itemSize, 0L );
+      dst += itemSize;
+      *dst++ = '\n'; // add our element seperator
+    }
+    ret = GetFlavorFlags( dragRef, itemRef, 'hfs ', &flags );
+    if ( ret == noErr )
+    {
+      HFSFlavor hfs; itemSize = sizeof( hfs );
+      GetFlavorData( dragRef, itemRef, 'hfs ', &hfs, &itemSize, 0L );
+      itemSize = FSSpec2UnixPath( &hfs.fileSpec, dst );
+      dst += itemSize;
+      if ( itemSize>1 && ( hfs.fileType=='fold' || hfs.fileType=='disk' ) ) 
+        *dst++ = '/';
+      *dst++ = '\n'; // add our element seperator
+    }
+  }
+
+  dst[-1] = 0;
+  currDragSize = dst - currDragData - 1;
+  currDragErr = ret;
+  return ret;
+}
+
 /**
  * Drag'n'drop tracking handler
  */
@@ -1376,6 +1569,10 @@ static pascal OSErr dndTrackingHandler( DragTrackingMessage msg, WindowPtr w, vo
   Point mp;
   static int px, py;
   
+  fillCurrentDragData(dragRef);
+  Fl::e_length = currDragSize;
+  Fl::e_text = currDragData;
+  
   switch ( msg )
   {
   case kDragTrackingEnterWindow:
@@ -1439,67 +1636,27 @@ static pascal OSErr dndReceiveHandler( WindowPtr w, void *userData, DragReferenc
   Fl::e_y = Fl::e_y_root - target->y();
   if ( !Fl::handle( FL_DND_RELEASE, target ) )
     return userCanceledErr;
-    
-  // get the ASCII text
-  UInt16 i, nItem;
-  ItemReference itemRef;
-  FlavorFlags flags;
-  Size itemSize, size = 0;
-  CountDragItems( dragRef, &nItem );
-  for ( i = 1; i <= nItem; i++ )
-  {
-    GetDragItemReferenceNumber( dragRef, i, &itemRef );
-    ret = GetFlavorFlags( dragRef, itemRef, 'TEXT', &flags );
-    if ( ret == noErr )
-    {
-      GetFlavorDataSize( dragRef, itemRef, 'TEXT', &itemSize );
-      size += itemSize;
-    }
-    ret = GetFlavorFlags( dragRef, itemRef, 'hfs ', &flags );
-    if ( ret == noErr )
-    {
-      size += 1024; //++ ouch! We should create the full pathname and figure out its length
-    }
-  }
-  
-  if ( !size )
+
+  ret = fillCurrentDragData(dragRef);
+  if (ret==userCanceledErr)
     return userCanceledErr;
   
-  Fl::e_length = size + nItem - 1;
-  char *dst = Fl::e_text = (char*)malloc( size+nItem );;
-  
-  for ( i = 1; i <= nItem; i++ )
-  {
-    GetDragItemReferenceNumber( dragRef, i, &itemRef );
-    ret = GetFlavorFlags( dragRef, itemRef, 'TEXT', &flags );
-    if ( ret == noErr )
-    {
-      GetFlavorDataSize( dragRef, itemRef, 'TEXT', &itemSize );
-      GetFlavorData( dragRef, itemRef, 'TEXT', dst, &itemSize, 0L );
-      dst += itemSize;
-      *dst++ = '\n'; // add our element seperator
-    }
-    ret = GetFlavorFlags( dragRef, itemRef, 'hfs ', &flags );
-    if ( ret == noErr )
-    {
-      HFSFlavor hfs; itemSize = sizeof( hfs );
-      GetFlavorData( dragRef, itemRef, 'hfs ', &hfs, &itemSize, 0L );
-      itemSize = FSSpec2UnixPath( &hfs.fileSpec, dst );
-      dst += itemSize;
-      if ( itemSize>1 && ( hfs.fileType=='fold' || hfs.fileType=='disk' ) ) 
-        *dst++ = '/';
-      *dst++ = '\n'; // add our element seperator
-    }
-  }
-  
-  dst[-1] = 0;
-//  if ( Fl::e_text[Fl::e_length-1]==0 ) Fl::e_length--; // modify, if trailing 0 is part of string
-  Fl::e_length = dst - Fl::e_text - 1;
+  Fl::e_length = currDragSize;
+  Fl::e_text = currDragData;
 //  printf("Sending following text to widget %p:\n%s\n", Fl::belowmouse(), Fl::e_text);
-  target->handle(FL_PASTE);
-  free( Fl::e_text );
-
+  int old_event = Fl::e_number;
+  Fl::belowmouse()->handle(Fl::e_number = FL_PASTE);
+  Fl::e_number = old_event;
+  
+  if (currDragData) {
+    free(currDragData);
+  }
+  currDragData = 0L;
+  currDragRef = 0;
+  Fl::e_text = 0L;
+  Fl::e_length = 0;
   fl_dnd_target_window = 0L;
+  
   breakMacEventLoop();
   return noErr;
 }
@@ -1539,7 +1696,9 @@ void Fl_X::make(Fl_Window* w)
       x->wait_for_expose = 0;
       x->next = Fl_X::first; // must be in the list for ::flush()
       Fl_X::first = x;
-      w->handle(FL_SHOW);
+      int old_event = Fl::e_number;
+      w->handle(Fl::e_number = FL_SHOW);
+      Fl::e_number = old_event;
       w->redraw(); // force draw to happen
     }
     fl_show_iconic = 0;
@@ -1572,8 +1731,12 @@ void Fl_X::make(Fl_Window* w)
 
     if (!fake_X_wm(w, xwm, ywm, bt, bx, by)) {
       // menu windows and tooltips
-      winclass = kHelpWindowClass;
-      winattr  = 0;
+      if (w->modal()||w->override()) {
+        winclass = kHelpWindowClass;
+        winattr  = 0;
+      } else {
+        winattr = 512; // kWindowNoTitleBarAttribute;
+      }
     } else if (w->modal()) {
       winclass = kMovableModalWindowClass;
     }
@@ -1650,14 +1813,15 @@ void Fl_X::make(Fl_Window* w)
     x->wait_for_expose = 1;
     x->next = Fl_X::first;
     Fl_X::first = x;
-    if (w->resizable()) DrawGrowIcon(x->xid);
     w->set_visible();
     { // Install Carbon Event handlers 
       OSStatus ret;
       EventHandlerUPP mousewheelHandler = NewEventHandlerUPP( carbonMousewheelHandler ); // will not be disposed by Carbon...
       static EventTypeSpec mousewheelEvents[] = {
         { kEventClassMouse, kEventMouseWheelMoved } };
-      ret = InstallWindowEventHandler( x->xid, mousewheelHandler, 1, mousewheelEvents, w, 0L );
+      ret = InstallWindowEventHandler( x->xid, mousewheelHandler,
+	        (int)(sizeof(mousewheelEvents)/sizeof(mousewheelEvents[0])),
+		mousewheelEvents, w, 0L );
       EventHandlerUPP mouseHandler = NewEventHandlerUPP( carbonMouseHandler ); // will not be disposed by Carbon...
       static EventTypeSpec mouseEvents[] = {
         { kEventClassMouse, kEventMouseDown },
@@ -1708,7 +1872,9 @@ void Fl_X::make(Fl_Window* w)
     w->x(rect.left); w->y(rect.top);
     w->w(rect.right-rect.left); w->h(rect.bottom-rect.top);
 
-    w->handle(FL_SHOW);
+    int old_event = Fl::e_number;
+    w->handle(Fl::e_number = FL_SHOW);
+    Fl::e_number = old_event;
     w->redraw(); // force draw to happen
     w->set_visible();
     
@@ -1817,7 +1983,10 @@ void Fl_Window::resize(int X,int Y,int W,int H) {
   resize_from_system = 0;
   if (is_a_resize) {
     Fl_Group::resize(X,Y,W,H);
-    if (shown()) { redraw(); if (!parent()) i->wait_for_expose = 1; }
+    if (shown()) { 
+      redraw(); 
+      //if (!parent()) i->wait_for_expose = 1; 
+    }
   } else {
     x(X); y(Y); 
   }
@@ -1856,8 +2025,9 @@ void Fl_Window::make_current()
   // \todo for performance reasons: we don't have to create this unless the child windows moved
   for ( Fl_X *cx = i->xidChildren; cx; cx = cx->xidNext )
   {
-    Fl_Region r = NewRgn();
     Fl_Window *cw = cx->w;
+    if (!cw->visible_r()) continue;
+    Fl_Region r = NewRgn();
     SetRectRgn( r, cw->x() - xp, cw->y() - yp, 
                    cw->x() + cw->w() - xp, cw->y() + cw->h() - yp );
     DiffRgn( fl_window_region, r, fl_window_region );
@@ -2011,8 +2181,63 @@ void Fl::paste(Fl_Widget &receiver, int clipboard) {
   return;
 }
 
+void Fl::add_timeout(double time, Fl_Timeout_Handler cb, void* data)
+{
+    int timer_id = -1;
+    for (int i = 0;  i < mac_timer_used;  ++i) {
+        if ( !mac_timers[i].timer ) {
+            timer_id = i;
+            break;
+        }
+    }
+    if (timer_id == -1) {
+        if (mac_timer_used == mac_timer_alloc) {
+            realloc_timers();
+        }
+        timer_id = mac_timer_used++;
+    }
+
+    EventTimerInterval fireDelay = (EventTimerInterval) time;
+    EventLoopTimerUPP  timerUPP = NewEventLoopTimerUPP(do_timer);
+    EventLoopTimerRef  timerRef;
+    OSStatus err = InstallEventLoopTimer(GetMainEventLoop(), fireDelay, 0, timerUPP, data, &timerRef);
+    if (err == noErr) {
+        mac_timers[timer_id].callback = cb;
+        mac_timers[timer_id].data     = data;
+        mac_timers[timer_id].timer    = timerRef;
+    }
+}
+
+void Fl::repeat_timeout(double time, Fl_Timeout_Handler cb, void* data)
+{
+    remove_timeout(cb, data);
+    add_timeout(time, cb, data);
+}
+
+int Fl::has_timeout(Fl_Timeout_Handler cb, void* data)
+{
+   for (int i = 0;  i < mac_timer_used;  ++i) {
+        MacTimeout& t = mac_timers[i];
+        if (t.callback == cb  &&  t.data == data) {
+            return 1;
+        }
+    }
+    return 0;
+}
+
+void Fl::remove_timeout(Fl_Timeout_Handler cb, void* data)
+{
+   for (int i = 0;  i < mac_timer_used;  ++i) {
+        MacTimeout& t = mac_timers[i];
+        if (t.callback == cb  && ( t.data == data || data == NULL)) {
+            delete_timer(t);
+        }
+    }
+}
+
+
 
 //
-// End of "$Id: Fl_mac.cxx,v 1.1.2.64 2004/09/09 21:34:46 matthiaswm Exp $".
+// End of "$Id: Fl_mac.cxx 4748 2006-01-15 02:26:54Z mike $".
 //
 
