@@ -1,11 +1,11 @@
 //
-// "$Id: editor.cxx,v 1.2.2.3.2.18 2004/04/11 04:39:01 easysw Exp $"
+// "$Id: editor.cxx 4723 2005-12-30 10:13:17Z matt $"
 //
 // A simple text editor program for the Fast Light Tool Kit (FLTK).
 //
 // This program is described in Chapter 4 of the FLTK Programmer's Guide.
 //
-// Copyright 1998-2004 by Bill Spitzak and others.
+// Copyright 1998-2005 by Bill Spitzak and others.
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Library General Public
@@ -22,7 +22,9 @@
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 // USA.
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+// Please report all bugs and problems on the following page:
+//
+//     http://www.fltk.org/str.php
 //
 
 //
@@ -150,13 +152,14 @@ const char         *code_types[] = {	// List of known C/C++ types...
 // 'compare_keywords()' - Compare two keywords...
 //
 
-int
-compare_keywords(const void *a,
-                 const void *b) {
-  return (strcmp(*((const char **)a), *((const char **)b)));
+extern "C" {
+  int
+  compare_keywords(const void *a,
+                   const void *b) {
+    return (strcmp(*((const char **)a), *((const char **)b)));
+  }
 }
 
-
 //
 // 'style_parse()' - Parse text and produce style data.
 //
@@ -172,8 +175,18 @@ style_parse(const char *text,
              *bufptr;
   const char *temp;
 
+  // Style letters:
+  //
+  // A - Plain
+  // B - Line comments
+  // C - Block comments
+  // D - Strings
+  // E - Directives
+  // F - Types
+  // G - Keywords
+
   for (current = *style, col = 0, last = 0; length > 0; length --, text ++) {
-    if (current == 'B') current = 'A';
+    if (current == 'B' || current == 'F' || current == 'G') current = 'A';
     if (current == 'A') {
       // Check for directives, comments, strings, and keywords...
       if (col == 0 && *text == '#') {
@@ -196,13 +209,13 @@ style_parse(const char *text,
 	continue;
       } else if (*text == '\"') {
         current = 'D';
-      } else if (!last && islower(*text)) {
+      } else if (!last && (islower(*text) || *text == '_')) {
         // Might be a keyword...
 	for (temp = text, bufptr = buf;
-	     islower(*temp) && bufptr < (buf + sizeof(buf) - 1);
+	     (islower(*temp) || *temp == '_') && bufptr < (buf + sizeof(buf) - 1);
 	     *bufptr++ = *temp++);
 
-        if (!islower(*temp)) {
+        if (!islower(*temp) && *temp != '_') {
 	  *bufptr = '\0';
 
           bufptr = buf;
@@ -271,7 +284,7 @@ style_parse(const char *text,
     else *style++ = current;
     col ++;
 
-    last = isalnum(*text) || *text == '.';
+    last = isalnum(*text) || *text == '_' || *text == '.';
 
     if (*text == '\n') {
       // Reset column and possibly reset the style
@@ -357,28 +370,35 @@ style_update(int        pos,		// I - Position of update
   stylebuf->select(pos, pos + nInserted - nDeleted);
 
   // Re-parse the changed region; we do this by parsing from the
-  // beginning of the line of the changed region to the end of
+  // beginning of the previous line of the changed region to the end of
   // the line of the changed region...  Then we check the last
   // style character and keep updating if we have a multi-line
   // comment character...
   start = textbuf->line_start(pos);
+//  if (start > 0) start = textbuf->line_start(start - 1);
   end   = textbuf->line_end(pos + nInserted);
   text  = textbuf->text_range(start, end);
   style = stylebuf->text_range(start, end);
-  last  = style[end - start - 1];
+  if (start==end)
+    last = 0;
+  else
+    last  = style[end - start - 1];
 
-//  printf("start = %d, end = %d, text = \"%s\", style = \"%s\"...\n",
-//         start, end, text, style);
+//  printf("start = %d, end = %d, text = \"%s\", style = \"%s\", last='%c'...\n",
+//         start, end, text, style, last);
 
   style_parse(text, style, end - start);
 
-//  printf("new style = \"%s\"...\n", style);
+//  printf("new style = \"%s\", new last='%c'...\n", 
+//         style, style[end - start - 1]);
 
   stylebuf->replace(start, end, style);
   ((Fl_Text_Editor *)cbArg)->redisplay_range(start, end);
 
-  if (last != style[end - start - 1]) {
-    // The last character on the line changed styles, so reparse the
+  if (start==end || last != style[end - start - 1]) {
+//    printf("Recalculate the rest of the buffer style\n");
+    // Either the user deleted some text, or the last character 
+    // on the line changed styles, so reparse the
     // remainder of the buffer...
     free(text);
     free(style);
@@ -453,7 +473,7 @@ int check_save(void) {
 
   int r = fl_choice("The current file has not been saved.\n"
                     "Would you like to save it now?",
-                    "Cancel", "Save", "Discard");
+                    "Cancel", "Save", "Don't Save");
 
   if (r == 1) {
     save_cb(); // Save the file...
@@ -771,5 +791,5 @@ int main(int argc, char **argv) {
 }
 
 //
-// End of "$Id: editor.cxx,v 1.2.2.3.2.18 2004/04/11 04:39:01 easysw Exp $".
+// End of "$Id: editor.cxx 4723 2005-12-30 10:13:17Z matt $".
 //
